package client;

/* This file is Copyright by Walter W. Schilling, Jr. in the year 2004.
 * All rights of usage are limited. */

/**
 * @author Walter Schilling
 * $Author: wws $
 * $Revision: 1.9 $
 * $Date: 2004/04/27 22:49:56 $
 * $Header: /cvs/java_lab6/ChatGUIPanel.java,v 1.9 2004/04/27 22:49:56 wws Exp $
 * $Name:  $
 * This class implements the main panel for the Simple text chat client.  
 * This includes the entire GUI, etc.
 * 
 * $Log: ChatGUIPanel.java,v $
 * Revision 1.9  2004/04/27 22:49:56  wws
 * Updated javadoc representation of routines / methods to be consistent and relativly correct.
 *
 * Revision 1.8  2004/04/27 22:20:49  wws
 * Added support to shutdown the remote server when the last user disconnects from the server.
 * Performed minor refactoring of code.
 *
 * Revision 1.7  2004/04/20 22:18:26  wws
 * Implemented RMI methodology for accessing chat server.
 *
 * Revision 1.6  2004/04/18 13:35:08  wws
 * Changed interface so it naturally deals with rmi exceptions.
 *
 * Revision 1.5  2004/04/17 11:07:21  wws
 * Added capability to save the textual log of the chat session to a file.
 *
 * Revision 1.4  2004/04/17 10:56:09  wws
 * Modified GUI so that it operated in a polled, multithreaded manner.
 *
 * Revision 1.3  2004/04/16 21:30:45  wws
 * Added simplistic transmission and reception, as well as adding new button to spawn a second gui.  Transmissions can now occur between GUIs using basic chat protocol.
 *
 * Revision 1.2  2004/04/13 21:34:10  wws
 * Updated GUI to include support for enabling and disabling of the menu and lower buttons when conected and disconnected.
 *
 * Revision 1.1  2004/04/13 21:05:49  wws
 * Initial Chat Client GUI development.
 *
 *
 */

import common.ChatTextualEntry;
import common.ChatMessageList;
import common.ChatUserInterface;
import common.DebugLog;
import common.DuplicateUserException;
import common.MessageReceivedCallbackHandler;

import server.ChatServerClientInterface;
import java.util.*;
import java.io.FileWriter;
import java.io.IOException;
import java.rmi.*;

public class ChatInfrastructure implements Runnable {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private ClientUser user = new ClientUser("", "");
	private String server = "";
	private ChatServerClientInterface l_interface = null;
	private Date l_last_timestamp;
	private boolean l_continue_thread_running;
	private Thread l_my_thread = new Thread(this);
	private MessageReceivedCallbackHandler callbackHandler;
	private ChatMessageList messageList = null;
	private ChatMessageList unblockedMessages = null;
	private boolean connectedToServer = false;

	private Hashtable<String, ClientUser> userTable = new Hashtable<String, ClientUser>();

	/**
	 * This constructor will construct a ChatGUIPanel.
	 * 
	 * @param t_interface
	 *            This si the interface for performing chat communications that
	 *            is to be used.
	 */
	public ChatInfrastructure() {
		l_continue_thread_running = false;
		unblockedMessages = new ChatMessageList();
	}
	
	public void finalize()
	{
		
	}

	/**
	 * This method will clear all messages from the logs.
	 */
	public void clearMessages()
	{
		messageList.removeAll(messageList);
		messageList.clear();
		
	}
	
	
	/**
	 * This method will invoke the thread which is created for the panel. It is
	 * responsible for querying the text server on a periodic basis to retrieve
	 * new messages.
	 */
	public void run() {
		while (l_continue_thread_running == true) {
			try {
				// Note: Users table must be updated before the new text can be
				// obtained from another server. Otherwise, there are problems
				// with new users text not showing up because they are not yet
				// a valid user in the system, resulting in problems with the
				// ignored state ignoring the text from new users. Thus, this
				// sequence must be kept.
				updateUsersTable();
				obtainNewTextFromServer();
				updateHeartbeatWithServer();
				Thread.sleep(1000);

			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		// If we get here, the thread is shutting down. So, simply create a new
		// instance to be ready to run again.
		this.l_my_thread = new Thread(this);
	}

	private void updateHeartbeatWithServer() {
		try {
			this.l_interface.checkinWithServer(this.user);
		} catch (RemoteException e) {
			this.handleConnectionError();
		}
	}

	/**
	 * This method is responsible for updating the text panel with new textual
	 * strings. Due to the multithreaded nature, this must be a thread safe
	 * routine, and thus, the synchronized definition.
	 */
	private synchronized void obtainNewTextFromServer() {
		// Start by getting any new text messages from the server.
		ChatMessageList newMessages;
		try {
			newMessages = this.l_interface
					.obtain_new_text_strings(this.l_last_timestamp);
			// Reset the timestamp for the last update to the most recent
			// timestamp.
			this.l_last_timestamp = newMessages.get_timestamp();

			// Check to see if the number of messages received is not 0.
			if (newMessages.size() > 0) {

				// Add all the messages to the local message cache.
				this.messageList.addAll(newMessages);

				// Process the messages received to see if there are any from
				// unblocked users. If there are, then update the display.
				if (processNonIgnoredMessages(newMessages) == true) {
					callbackHandler.newMessageReceived();
				}
			}
		} catch (RemoteException e) {
		}
	}

	/**
	 * @param newMessages
	 * @return true if there is at least one message from an unignored-user
	 *         which has been received. false otherwise.
	 */
	private boolean processNonIgnoredMessages(ChatMessageList newMessages) {
		boolean returnValue = false;
		// Now add the unblocked messages to the listing of unblocked messages.
		for (ChatTextualEntry message : newMessages) {
			ClientUser userID = this.userTable.get(message.get_username());
			if (userID != null) {
				// Check to see if the user is blocked.
				if (userID.isIgnored() == true) {
					// Do nothing with this message, as the user is blocked.
				} else {
					// Add it to the listing
					this.unblockedMessages.add(message);
					returnValue = true;
				}
			}
		}
		return returnValue;
	}

	private void handleConnectionError() {
		// TODO Implement this.

	}

	/**
	 * This method will be called when the user wishes to save the chat contents
	 * into a file.
	 * 
	 * @param p_filename
	 *            This is the filename that shall be created.
	 */
	public void export_textlog(FileWriter writer) throws IOException {
		for (ChatTextualEntry msg : this.unblockedMessages) {
			writer.write(msg.toString());
		}
	}

	/**
	 * This method will be invoked when the user selects the disconnect menu
	 * option.
	 */
	public void disconnect_from_server() {
		DebugLog debugLog = new DebugLog();
		debugLog.println("Disbling things...");
		// Check the connected state. If we are not connected, then we really do
		// not care.

		if (connectedToServer == true) {
			ChatMessageList t_new_text;
			try {
				t_new_text = l_interface.logoff_user(user);
				this.messageList.addAll(t_new_text);
				callbackHandler.newMessageReceived();

				if (l_interface.getUserCount() == 0) {
					/*
					 * Now as one final thing, lets see if anyone else is
					 * connected. If not, see if the user wants to shut down the
					 * server.
					 */
					if (this.callbackHandler.requestServerLogoff() == true) {
						try {
							l_interface.shutdown_server();
						} catch (RemoteException exp) {
							exp.printStackTrace();
						}
					}

				}
				l_interface = null;
			} catch (RemoteException exp) {
				// TODO Handle this properly.
				t_new_text = new ChatMessageList();
				exp.printStackTrace();
			}

			l_last_timestamp = t_new_text.get_timestamp();
			l_continue_thread_running = false;
			connectedToServer = false;
		}
	}

	/**
	 * This method will be invoked when the user invokes the connect menu
	 * option.
	 */
	public boolean connectWithServer(String serverName, String username,
			String screenName) throws NotBoundException, RemoteException,
			ClassCastException, DuplicateUserException, Exception {
		String t_registry = serverName;
		String t_registration = "rmi://" + t_registry + "/RMIChat1";
		DebugLog.println("Connecting to... " + t_registration);
		Remote t_remote_service = Naming.lookup(t_registration);

		DebugLog.println(t_remote_service.toString());
		this.l_interface = (ChatServerClientInterface) Naming
				.lookup(t_registration);

		this.server = serverName;
		this.user = new ClientUser(username, screenName);
		this.callbackHandler.setServerName(this.server);
		this.callbackHandler.setUserName(this.user.getUserName() + " <"
				+ this.user.getUserScreenName() + ">");

		messageList = l_interface.login_user(user);
		this.updateUsersTable();
		processNonIgnoredMessages(messageList);
		l_last_timestamp = messageList.get_timestamp();

		// Start the thread running.
		l_continue_thread_running = true;
		l_my_thread.start();
		callbackHandler.newMessageReceived();
		callbackHandler.setServerName(this.server);
		callbackHandler.indicateConnectionEstabolished(server);
		connectedToServer = true;
		return true;
	}

	/**
	 * This method will be invoked when the user presses the transmit button,
	 * causing a text string to be transmitted to the server.
	 * 
	 */
	public boolean transmitText(String textToTransmit) {
		/*
		 * This event handler will handle a request to transmit text to other
		 * users of the chat room.
		 */
		ChatTextualEntry t_textual_entry = new ChatTextualEntry(user,
				textToTransmit);
		DebugLog.println(t_textual_entry.toString());
		boolean returnValue = false;

		try {
			l_interface.send_chat_text(t_textual_entry);
			returnValue = true;
		} catch (RemoteException exp) {
			// TODO Fix this properly.
			exp.printStackTrace();
		}

		this.obtainNewTextFromServer();
		return returnValue;
	}

	private void updateUsersTable() {
		// Get the users listing from the server.
		Collection<ChatUserInterface> serverUserList;
		try {
			serverUserList = this.l_interface.getUserList();

			boolean listChanged = false;

			for (ChatUserInterface user : serverUserList) {
				if (this.userTable.containsKey(user.getUserScreenName()) == false) {
					listChanged = true;
					this.userTable.put(user.getUserScreenName(),
							new ClientUser(user));
				}

			}

			Set<String> keys = this.userTable.keySet();
			ChatUserInterface[] returnArray = new ChatUserInterface[serverUserList
					.size()];

			returnArray = (ChatUserInterface[]) serverUserList
					.toArray(returnArray);

			ArrayList<String> keysToRemove = new ArrayList<String>();

			for (String currentKey : keys) {
				boolean t_match_found = false;
				int index = 0;

				while ((index < returnArray.length) && (t_match_found == false)) {
					if (returnArray[index].getUserScreenName().equals(
							currentKey)) {
						t_match_found = true;
					}
					index++;
				}
				if (t_match_found == false) {
					// The user is no longer connected. Remove from the cached
					// list
					// of users.
					keysToRemove.add(currentKey);
					listChanged = true;
				}
			}

			for (String currentKey : keysToRemove) {
				this.userTable.remove(currentKey);
			}

			// Now that the lists are synchronized, do an update if we need to.
			if (listChanged == true) {
				this.callbackHandler.indicateUserChange();
			}
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public MessageReceivedCallbackHandler getCallbackHandler() {
		return callbackHandler;
	}

	public void setCallbackHandler(
			MessageReceivedCallbackHandler callbackHandler) {
		this.callbackHandler = callbackHandler;
	}

	public ChatMessageList getMessageList() {
		return messageList;
	}

	public void setMessageList(ChatMessageList messageList) {
		this.messageList = messageList;
	}

	public Hashtable<String, ClientUser> getUserTable() {
		return userTable;
	}

	public void setUserTable(Hashtable<String, ClientUser> userTable) {
		this.userTable = userTable;
	}

	public ArrayList<ClientUser> getUsers() {

		return new ArrayList<ClientUser>(this.userTable.values());
	}

	public ChatMessageList getUnblockedMessages() {
		return unblockedMessages;
	}

	public void setUnblockedMessages(ChatMessageList unblockedMessages) {
		this.unblockedMessages = unblockedMessages;
	}

	public ClientUser getUser() {
		return user;
	}

	public void setUser(ClientUser user) {
		this.user = user;
	}

	public String getServer() {
		return server;
	}

	public void setServer(String server) {
		this.server = server;
	}

	public void autoConnect() {
		try {
			this.connectWithServer(this.server, this.user.getUserName(),
					this.user.getUserScreenName());
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassCastException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotBoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (DuplicateUserException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

}