package client.ui;

import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import client.ChatInfrastructure;

import common.ChatMessageList;
import common.DebugLog;
import common.DuplicateUserException;
import common.MessageReceivedCallbackHandler;

public class ChatGUIController implements MessageReceivedCallbackHandler {
	private ChatGUIPanel gui = null;
	private ChatInfrastructure infrastructure = null;
	private String leftTitle = "";
	private String rightTitle = "";
	private boolean playSound = true;

	public void quit() {
		this.disconnectFromServer();
		System.exit(0);
	}

	public ChatGUIPanel getGui() {
		return gui;
	}

	public void setGui(ChatGUIPanel gui) {
		this.gui = gui;
	}

	public void saveMessageLog() {
		String t_filename_to_export_to = this.gui.getMessageLogFilename();
		FileWriter writer = null;
		try {
			writer = new FileWriter(t_filename_to_export_to);
			this.infrastructure.export_textlog(writer);
		} catch (IOException e) {

			gui.showErrorMessage("Unable to write file "
					+ t_filename_to_export_to + ".", "File Save Error!");
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (IOException e) {
					gui
							.showErrorMessage("Unable to write file "
									+ t_filename_to_export_to + ".",
									"File Save Error!");
				}
			}
		}
	}

	public void transmitMessage(String text) {
		boolean successful = this.infrastructure.transmitText(text);
		if (successful == true) {
			this.gui.clearTextToTransmitEntry();
		}
	}

	public void disconnectFromServer() {
		DebugLog.println("Disbling things...");
		this.gui.updateConnectionState(false);

		this.infrastructure.disconnect_from_server();
	}

	public void connectToServer() {
		// Obtain the Server name to connect with.
		String serverName = gui.getServerName(this.infrastructure.getServer());

		// Obtain the User Name to display on the screen.
		String user = gui.getUserName(this.infrastructure.getUser()
				.getUserName());

		String screenName = gui.getScreenName(this.infrastructure.getUser()
				.getUserScreenName());

		try {
			infrastructure.connectWithServer(serverName, user, screenName);
		} catch (NotBoundException notBound) {
			gui.showErrorMessage("Unable to connect with server " + serverName
					+ ".", "Connection Error!");
			DebugLog.println("The server could not be found.");
		} catch (RemoteException remEx) {
			gui.showErrorMessage(
					"Remote exception occurred on the remote server "
							+ serverName + ".", "Connection Error!");
			DebugLog
					.println("A remote exception has occurred while communicating with the server.");
		} catch (MalformedURLException ex) {
			gui.showErrorMessage(
					"Improper URL provided for hostname.  Unable to connect. ",
					"Malforned URL");
			DebugLog.println("Malformed URL.");

		}

		catch (ClassCastException ex) {
			DebugLog.println("A class casting exception has occurred.");
			ex.printStackTrace();
			DebugLog.println(ex.getLocalizedMessage());
			DebugLog.println(ex.getMessage());
			DebugLog.println(ex.toString());
		} catch (DuplicateUserException e) {
			gui
					.showErrorMessage(
							"User "
									+ e.getUser().getUserScreenName()
									+ "is already logged onto the remote machine.  Can not duplicate login.",
							"Duplicate Login Attempted");
			DebugLog.println("Duplicate login attempted.");
		} catch (Exception e) {
			DebugLog.println(e.toString());
			DebugLog.println("A generic exception has occurred.");
			e.printStackTrace();
		}
	}

	public ChatInfrastructure getInfrastructure() {
		return infrastructure;
	}

	public void setInfrastructure(ChatInfrastructure infrastructure) {
		this.infrastructure = infrastructure;
	}

	public void newMessageReceived() {
		// Start by obtaining the message listing from the infrastructure.
		ChatMessageList messages = this.infrastructure.getUnblockedMessages();
		String output = messages.toString();
		this.gui.update_textbox_with_new_text(output);
		if (this.playSound == true) {
			new WavePlayer("beep.wav").start();
		}
	}

	@Override
	public void indicateConnectionEstabolished(String server) {
		this.gui.updateConnectionState(true);
	}

	@Override
	public boolean requestServerLogoff() {
		return this.gui.showServerShutdownDisplay();
	}

	@Override
	public void indicateUserChange() {
		this.gui.updateUsersListing(this.infrastructure.getUsers());
	}

	@Override
	public void setUserName(String user) {
		this.leftTitle = user;
		this.updatePanelTitle();
	}

	@Override
	public void setServerName(String server) {
		this.rightTitle = server;
		this.updatePanelTitle();
	}

	private void updatePanelTitle() {
		this.gui.setMyTitle(this.leftTitle + "@" + this.rightTitle);
	}

	public void CreateNewProgramInstance() {
		new ChatGUIProgramInstance(this.infrastructure.getServer(),
				this.infrastructure.getUser());
	}

	public boolean isPlaySound() {
		return playSound;
	}

	public void setPlaySound(boolean playSound) {
		this.playSound = playSound;
	}

}
