package server;

/* This file is Copyright by Walter W. Schilling, Jr. in the year 2004.
 * All rights of usage are limited. */

/**
 * @author Walter Schilling
 * $Author: wws $
 * $Revision: 1.4 $
 * $Date: 2004/04/27 22:49:56 $
 * $Header: /cvs/java_lab6/LocalJavaServer.java,v 1.4 2004/04/27 22:49:56 wws Exp $
 * $Name:  $
 * This is a local Java Server for the Java Chat system.
 * 
 * $Log: LocalJavaServer.java,v $
 * Revision 1.4  2004/04/27 22:49:56  wws
 * Updated javadoc representation of routines / methods to be consistent and relativly correct.
 *
 * Revision 1.3  2004/04/27 22:20:49  wws
 * Added support to shutdown the remote server when the last user disconnects from the server.
 * Performed minor refactoring of code.
 *
 * Revision 1.2  2004/04/20 22:18:26  wws
 * Implemented RMI methodology for accessing chat server.
 *
 * Revision 1.1  2004/04/16 21:30:04  wws
 * Initial implementation of server and server interface.  Server currently runs locally as of right now.
 *
 *
 */

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map.Entry;

import common.ChatTextualEntry;
import common.ChatMessageList;
import common.ChatUserInterface;
import common.DuplicateUserException;

public class LocalJavaServer implements ChatServerClientInterface, Runnable {
	private Hashtable<String, ServerUser> onlineUsers = new Hashtable<String, ServerUser>();

	private ArrayList<ChatTextualEntry> l_textual_transmissions;
	private Thread l_my_thread = new Thread(this);
	private boolean continueThreadRunning = true;

	private static final int HEARTBEAT_COUNT = -1;
	private static final int DISCONNECT_COUNT = -100;

	/**
	 * This method will add a new user into the server.
	 * 
	 * @param p_username
	 *            This is the user logging onto the system.
	 */
	public ChatMessageList login_user(ChatUserInterface p_username)
			throws DuplicateUserException {
		Date t_now = new Date();
		ChatMessageList t_return_value = new ChatMessageList();
		ArrayList<ChatTextualEntry> t_temp_array_list = new ArrayList<ChatTextualEntry>();

		if (onlineUsers.containsKey(p_username.getUserScreenName()) == false) {
			this.onlineUsers.put(p_username.getUserScreenName(),
					new ServerUser(p_username));
			ChatTextualEntry t_logon_text = new ChatTextualEntry(
					new Date(),
					p_username,
					("User " + p_username.getUserScreenName() + " has joined the chat channel"));
			l_textual_transmissions.add(t_logon_text);

			t_return_value.set_timestamp(t_now);
			t_temp_array_list.add(t_logon_text);
			t_return_value.set_chat_transmissions(t_temp_array_list);

			t_return_value.set_usercount(this.getUserCount());
		} else {
			System.out.println(p_username.getUserName()
					+ p_username.getUserScreenName());
			throw new DuplicateUserException("User "
					+ p_username.getUserScreenName() + " already exists.",
					p_username);
		}

		return t_return_value;
	}

	/**
	 * This method will rewmove a user from the server.
	 * 
	 * @param p_username
	 *            This is the user logging off the system.
	 */
	public ChatMessageList logoff_user(ChatUserInterface p_username) {
		Date t_now = new Date();
		ChatMessageList t_return_value = new ChatMessageList();
		ArrayList<ChatTextualEntry> t_temp_array_list = new ArrayList<ChatTextualEntry>();

		ChatTextualEntry t_logon_text = new ChatTextualEntry(
				t_now,
				p_username,
				("User " + p_username.getUserScreenName() + " has departed the chat channel"));
		l_textual_transmissions.add(t_logon_text);

		this.onlineUsers.remove(p_username.getUserScreenName());

		t_return_value.set_timestamp(t_now);
		t_temp_array_list.add(t_logon_text);
		t_return_value.set_chat_transmissions(t_temp_array_list);

		return t_return_value;
	}

	/**
	 * This method will return the textual entries that have occurred since the
	 * last query.
	 * 
	 * @param p_previous_timestamp
	 *            This is the timestamp of the last time that the client queried
	 *            the server for new textual entries.
	 */
	public ChatMessageList obtain_new_text_strings(Date p_previous_timestamp) {
		int t_index;
		boolean t_continue_looping = true;
		t_index = l_textual_transmissions.size() - 1;
		/*
		 * This loop will step back through the time stamps to find the first
		 * one which is not newer than the previous timestamp. Once that is
		 * accomplished, we will break out of the loop.
		 */
		while ((t_index > 0) && (t_continue_looping == true)) {
			if (((ChatTextualEntry) l_textual_transmissions.get(t_index))
					.get_entry_timestamp().after(p_previous_timestamp) == true) {
				t_index--;
			} else {
				t_continue_looping = false;
			}
		}
		/*
		 * Now that we have found the first value older than the timestamp, we
		 * will increment by one and return all newer values.
		 */
		ArrayList<ChatTextualEntry> t_array_list = new ArrayList<ChatTextualEntry>();
		t_index++;
		while (t_index < l_textual_transmissions.size()) {
			t_array_list.add(l_textual_transmissions.get(t_index));
			t_index++;
		}

		ChatMessageList t_return_value = new ChatMessageList();
		t_return_value
				.set_timestamp(((ChatTextualEntry) (l_textual_transmissions
						.get(l_textual_transmissions.size() - 1)))
						.get_entry_timestamp());
		t_return_value.set_chat_transmissions(t_array_list);
		t_return_value.set_usercount(this.getUserCount());
		return t_return_value;
	}

	/**
	 * This method will add a new textual transmission.
	 * 
	 * @param p_entry
	 *            This is the textual transmission that has just been sent.
	 */
	public void send_chat_text(ChatTextualEntry p_entry) {
		p_entry.set_entry_timestamp(new Date());
		l_textual_transmissions.add(p_entry);
	}

	/**
	 * This method will be called to shutdown the server. It will start a thread
	 * which will lead to the self destruction of the server.
	 */
	public void shutdown_server() {
		continueThreadRunning = false;
	}

	/**
	 * This method is the thread run class method. It will wait 1 second and
	 * then will cause the server to self-destruct (i.e. exit).
	 */
	public void run() {
		/* Wait 1 second and then shutdown the thread. */

		while (continueThreadRunning == true) {
			try {
				Thread.sleep(1000);
				checkHeartbeats();

			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		try {
			this.finalize();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * This method will initialize a LocalJavaServer.
	 */
	public LocalJavaServer() throws java.rmi.RemoteException {
		super();
		l_textual_transmissions = new ArrayList<ChatTextualEntry>();
		l_my_thread.start();
	}

	@Override
	public int getUserCount() {
		return this.onlineUsers.size();
	}

	@Override
	public Collection<ChatUserInterface> getUserList() {
		Collection<ChatUserInterface> returnValue = new ArrayList<ChatUserInterface>();

		for (ChatUserInterface user : this.onlineUsers.values()) {
			returnValue.add(user);
		}

		return returnValue;
	}

	@Override
	public void checkinWithServer(ChatUserInterface user) {
		// Start by finding the instance of the user.
		ServerUser me = this.onlineUsers.get(user.getUserScreenName());

		if (me != null) {
			me.setHeartbeatCheck(5);
		}
	}

	public void checkHeartbeats() {
		Iterator<Entry<String, ServerUser>> hashIterator = this.onlineUsers
				.entrySet().iterator();
		while (hashIterator.hasNext() == true) {
			Entry<String, ServerUser> next = hashIterator.next();
			ServerUser me = next.getValue();
			me.setHeartbeatCheck(me.getHeartbeatCheck() - 1);
			if (me.getHeartbeatCheck() == HEARTBEAT_COUNT) {
				// Heartbeat Error.
				ChatTextualEntry t_logon_text = new ChatTextualEntry(
						new Date(),
						me,
						("User " + me.getUserScreenName() + " is experiencing network connectivity problems."));
				l_textual_transmissions.add(t_logon_text);
			} else if (me.getHeartbeatCheck() == DISCONNECT_COUNT) {
				// User has gone away and there doesn't seem to be a recovery
				// for them.
				// Remove them from the listing.
				ChatTextualEntry t_logon_text = new ChatTextualEntry(
						new Date(),
						me,
						("User " + me.getUserScreenName() + " has unexpectedly left the channel."));
				l_textual_transmissions.add(t_logon_text);

				this.onlineUsers.remove(next.getKey());
			}
		}
	}
}
